//////////////////////////////////////////////////
// convexgeom.m
//////////////////////////////////////////////////

load "dim0.m";
load "dim1.m";
load "dim2.m";
load "dimn.m";

// Returns a random hyperplane h oriented such that the points lie in the
// (strictly) positive half-space, and h evaluates to a different value on each
// of the points. Note that we assume the points are all distinct, and lie in
// the positive orthant.
function random_sweeping_hyperplane(pts)
	L:=Dual(Universe(pts));
	r:=5;
	while true do
		h:=PrimitiveLatticeVector(L ! [Random(r) : i in [1..Dimension(L)]]);
		vals:={v * h : v in pts};
		if #vals eq #pts and not 0 in vals then
			return h;
		end if;
		r +:= 1;
	end while;
end function;

// Note that we assume that P has non-trivial tail cone, and is contained in
// the positive orthant.
function split_to_cones(P)
	// Is this already an integrally translated cone?
    if NumberOfVertices(P) eq 1 and IsIntegral(Representative(Vertices(P))) then
        return [P];
    end if;
	// First build the set of points to consider
	C:=InfinitePart(P);
	Q:=Polytope(Vertices(P) cat
					[v + u : v in Vertices(P), u in ZGenerators(C)]);
	pts:=Points(Q);
	// Find a sweeping hyperplane and use it to order the points
	h:=random_sweeping_hyperplane(pts);
	pts:=SetToSequence(pts);
	heights:=[v * h : v in pts];
	ParallelSort(~heights,~pts);
	// Start building the cones
	Cs:=[];
	for v in pts do
		if not &or[v in D : D in Cs] then
			Append(~Cs,v + C);
		end if;
	end for;
	return Cs;
end function;

// Analyse the infinite family P. At this point we may assume that P is a
// translated cone v + C, where v is a lattice point.
function analyse_family(P)
	// Get the 1-dimensional case out of the way first
    if Dimension(P) eq 1 then
	    finite_sol := [PowerSequence(Integers())|];
	    poly_sol := [];
        k:=Eltseq(Rays(InfinitePart(P))[1]);
		u:=Eltseq(Vertices(P)[1]);
        bool,sol:=one_dim_solution(k,u);
        if bool then
            finite_sol cat:= sol;
        else
            Append(~poly_sol,sol);
        end if;
		return finite_sol,poly_sol;
    end if;
    // Next we check whether there is a trivial reason why most members of P
	// are not quasismooth. We reject such families, keeping only the finitely
	// many members who are quasismooth.
    bool,finite := quasismooth_family_test(P);
    if not bool then
		// If we're here then there are only finitely many quasismooth members
        return finite,[];
    end if;
    // Now analyse certain 2-dimensional families - the func checks dim.
    bool,finite,fams := two_dim_solution(P);
    if bool then
        return finite,fams;
    end if;
	// If we've got to here we couldn't rule out P as an infinite family
	return [PowerSequence(Integers())|],[P];
end function;

// Given matrices A and and b of compatible sizes, returns the sequence of all
// integer points v satisfying the inhomogeneous linear system
//    Av = b   and
//    v_i > 0  for 1<=i<=n,
// subject to the requirement that v might be interpreted as X_{v_{n+1}}
// in P(v_1,...,v_n).
// Note: Well-formedness and quasismoothness are only employed to help reduce a
// potentially infinite collection of solutions to a finite one; you should NOT
// assume that every solution returned is well-formed and quasismooth. Results
// returned here may still fail elsewhere.

// This function returns two sequences: the first is a sequence of
// hypersurfaces X = [a1..an,d], and the second is a sequence of polyhedrons
// that require further analysis (no guarantee of anything: integral points,
// qsmooth, wellformed,...).
function solve_system(A,b)
    // Sanity check
    dim:=NumberOfColumns(A);
    error if dim eq 0,
        "solve_system: Matrix A must be non-empty";
    error if NumberOfRows(A) ne NumberOfRows(b),
        "solve_system: Matrices A and b must be of compatible sizes";
    error if Ncols(b) ne 1,
        "solve_system: Target matrix b must be a column vector";
    // Check whether any solutions to Av = b exist
    As:=RowSequence(A);
    Bs:=Eltseq(b);
    bool,u,K:=IsConsistent(Transpose(Matrix(As)),Matrix([Bs]));
    if not bool then
        return [PowerSequence(Integers())|],[];
    end if;
    u:=Eltseq(u);
    if Dimension(K) eq 0 then
        // This is a unique solution
        return zero_dim_solution(K,u),[];
    elif Dimension(K) eq 1 then
        // This is a line of solutions
        k:=Eltseq(Basis(K)[1]);
        bool,sol:=one_dim_solution(k,u);
        if bool then
            return sol,[];
        else
            return [PowerSequence(Integers())|],[sol];
        end if;
    end if;
    // Our attempt to get a quick integral solution from IsConsistent didn't
    // work, so now we build the full polyhedron of solutions, imposing
    // positivity and well-ordering (in our usual sense a1>=a2, etc) in
    // construction.
    Iminus:=IdentityMatrix(Integers(),dim);
    for i in [1..dim - 2] do
        Iminus[i,i + 1]:=-1;
    end for;
    W:=ZeroSequence(Integers(),dim - 1);
    W[#W]:=1;
    As cat:= RowSequence(-A) cat Prune(RowSequence(Iminus));
    Bs cat:= Eltseq(-b) cat W;
    P:=PolyhedronWithInequalities(As,Bs);   // a priori P has rational vxs.
    if IsPolytope(P) then
        us:=[PowerSequence(Integers()) | Eltseq(p) : p in Points(P)];
        return [Universe(us) | u : u in us |
                    &and[c gt 0 : c in u] and check_1_strata(u)],[];
    end if;
	// We split this into a collection of translated cones and analyse those
	Fs:=split_to_cones(P);
    finite_sol:={PowerSequence(Integers())|};
    poly_sol:=[];
	while #Fs ne 0 do
		F:=Fs[#Fs];
		Prune(~Fs);
		finite_F,poly_F:=analyse_family(F);
		finite_sol join:= SequenceToSet(finite_F);
		Fidx:=Index(poly_F,F);
		if Fidx eq 0 then
			Fs:=poly_F cat Fs;
		else
			Fs:=[poly_F[i] : i in [1..#poly_F] | i ne Fidx] cat Fs;
			Append(~poly_sol,F);
		end if;
	end while;
	// Return what we have
	return SetToSequence(finite_sol),poly_sol;
end function;
