//////////////////////////////////////////////////
// dim0.m
//////////////////////////////////////////////////

//////////////////////////////////////////////////
// Auxilliary functions to identify candidate 1-strata that may fail
// quasismoothness - this vastly reduces the number of cases to test fully
// later.

// Return e.g. [ {1},{2,3,4} ] if rows 2,3,4 of M are the same and 1 is not.
function partition_rows(M)
    mapping:=SetToSequence(SequenceToSet(Rows(M)));   
    idxs:=[Index(mapping,R) : R in Rows(M)];   
    return [{i : i in [1..#idxs] | idxs[i] eq j} : j in [1..#mapping]];
end function;

// Return a seq [ [i1,i2], [i3,i4],... ] of suspicious 1-strata in v
// by identifying maximal strata which have only a single tgt mono
// and then pulling the 1-strata out of those.
function candidate_bad_1_strata(v)
    bad_1_strata:=[];
    n:=#v;
    // Row i of vMat has 1s in positions that are tgt monos at P_i.
    vMat:=Matrix(Integers(),[[IsDivisibleBy(v[n] - v[j],v[i]) select 1 else 0 :
                                           j in [1..n - 1]] : i in [1..n - 1]]);
    for samerows in {idxs : idxs in partition_rows(vMat) | #idxs ge 2} do
        idx:=Representative(samerows);
        if &+Eltseq(vMat[idx]) eq 1 then
            bad_1_strata cat:= [SetToSequence(ij) : ij in Subsets(samerows,2)];
        end if;
    end for;
    return bad_1_strata;
end function;

function is_k_tgt_mono_on_ij(i,j,k,X)
    n:=#X;
    d:=X[n];
    if d - X[k] lt 0 then // var xk doesn't appear in the equation at all
        return false;
    end if;
    xi:=X[i];
    xj:=X[j];
    g,u,v:=XGCD(xi,xj);
    bool,ratio:=IsDivisibleBy(d - X[k],g);
    if not bool then
        return false;
    end if;
    ker_i:=xj div g;
    ker_j:=xi div g;
    if u lt 0 then
        ker_j *:= -1;
        min_multiplier:=-((u * ratio) div ker_i);	// > 0
        return v * ratio + min_multiplier * ker_j ge 0;
    else
        ker_i *:= -1;
        min_multiplier:=-((v * ratio) div ker_j);
        return u * ratio + min_multiplier * ker_i ge 0;
    end if;
end function;

// X = [a0..an,d] is a solution as usual.
// bad_1_strata = [ [1,2], [4,5] ] is seq of indxs of poss non-qsmooth strata.
// Return false if X is NOT quasismooth along any one of these strata.
// Otherwise return true (so X stands a chance of being quasismooth).
// We may assume (from usage below) quasismooth at all 0-strata
function test_one_stratum(X,bad_1_strata)
    n:=#X;
    rows:=[];
    for ij in bad_1_strata do
        i,j:=Explode(ij);
        if IsDivisibleBy(X[n],X[i]) or IsDivisibleBy(X[n],X[j]) then
            // The ij stratum isn't actually on X, so no problem.
            continue;
        end if;
        if IsDefined(rows,i) then
            rowi:=rows[i];
        else
            rowi:=[IsDivisibleBy(X[n] - X[k],X[i]) select 1 else 0 :
                                                                 k in [1..n-1]];
            rows[i]:=rowi;
        end if;
        if &+rowi ne 1 then	// can't be 0 since qsmooth at all 0-strata
            continue;
        end if;
        if IsDefined(rows,j) then
            rowj:=rows[j];
        else
            rowj:=[IsDivisibleBy(X[n] - X[k],X[j]) select 1 else 0 :
                                                                 k in [1..n-1]];
            rows[j]:=rowj;
        end if;
        if rowi eq rowj then
            indu:=Index(Eltseq(rowi),1);
            if not &or[is_k_tgt_mono_on_ij(i,j,k,X) :
                                               k in [1..n - 1] | k ne indu] then
                return false;
            end if;
        end if;
    end for;
    return true;
end function;

// Return true if u is ordered like [ 5,4,3,2,1, * ].
function is_ordered(u)
    return &and[u[i] ge u[i + 1] : i in [1..#u - 2]];
end function;

function check_1_strata(u)
    n:=#u;
    return &and[u[i] ge u[i + 1] : i in [1..n - 2]] and
           &or[IsDivisibleBy(u[n] - u[i],u[n - 1]) : i in [1..n - 1]] and
           test_one_stratum(u,candidate_bad_1_strata(u));
end function;

