//////////////////////////////////////////////////
// dim1.m
//////////////////////////////////////////////////

// Input: u and k both unordered, so the result [all poss u+L*k] will be finite.
// Returns that seq.
// u is the first point in the strictly positive orthant in direction k. Method
// is to find upper and lower bounds for L.
function get_unordered(u,k)
    lower:=0;
    upper:=false;
    for i in [1..#u - 2] do
        if k[i] eq k[i + 1] then
            if u[i] lt u[i + 1] then	// Will never have solutions
                return [PowerSequence(Integers())|];
                // Else	no conditions imposed
            end if;
        elif k[i] gt k[i + 1] then
            lower:=Maximum(lower,Ceiling((u[i + 1] - u[i]) / (k[i] - k[i +1])));
        else // k[i] lt k[i + 1]
            uu:=Floor((u[i + 1] - u[i]) / (k[i] - k[i + 1]));
            if Type(upper) eq RngIntElt then
                upper:=Minimum(upper,uu);
            else
                upper:=uu;
            end if;
        end if;
        if Type(upper) eq RngIntElt and lower gt upper then
            return [PowerSequence(Integers())|];
        end if;
    end for;
    // Note: Upper MUST have been set since k IS unordered.
    res:=[PowerSequence(Integers())|];
    for a in [lower..upper] do
        v:=[Integers() | u[i] + a * k[i] : i in [1..#k]];
        if check_1_strata(v) then
            Append(~res,v);
        end if;
    end for;
    return res;
end function;

// Input: u is unordered but k is ordered.
// Output: First multiple u + L * k that is ordered.
function find_better_start(u,k)
    lower:=0;
    for i in [1..#u - 2] do
        if k[i] eq k[i + 1] then
            if u[i] lt u[i + 1] then	// Will never have solutions
                return false,_;
            end if;
        else	// k[i] gt k[i + 1] then
            lower:=Maximum(lower,Ceiling((u[i + 1] - u[i]) / (k[i] - k[i +1])));
        end if;
    end for;
    return true,[Integers() | u[i] + lower * k[i] : i in [1..#k]];
end function;

forward shift_point, first_qsmooth, second_qsmooth, most_not_wf, check_1_dim_proportional_strata, first_dim1_wellformed, second_dim1_wellformed;

function zero_dim_solution(u);
    if &and[c gt 0 : c in u] and is_ordered(u) then
        return [PowerSequence(Integers()) | u];
    else
        return [PowerSequence(Integers())|];
    end if;
end function;

function one_dim_solution(k,u);
    // First we need to shift u so that it's the smallest integral point in the
    // strictly positive orthant. At the same time, we detect whether this is a
    // bounded region or not.
    bool,obj:=shift_point(u,k);
    if bool then
        return true,[Universe(obj) | u : u in obj | check_1_strata(u)];
    end if;
    u:=obj[1];
    k:=obj[2];
    // Note: We only want correctly ordered solutions so work on that here.
    if is_ordered(u) then
        if not is_ordered(k) then
            result:=[u];
            u:=[Integers() | u[i] + k[i] : i in [1..#k]];
            while is_ordered(u) do
                if check_1_strata(u) then
                    Append(~result,u);
                end if;
                u:=[Integers() | u[i] + k[i] : i in [1..#k]];
            end while;
            return true,result;
        end if;
    else
        if is_ordered(k) then	// will be infinite - big k dominates
            bool,u:=find_better_start(u,k);
            if not bool then
                return true,[PowerSequence(Integers())|];
            end if;
        else	// will be finite
            result:=get_unordered(u,k);
            return true,result;
        end if;
    end if;
    // Next we check that there is no cheap obstruction to infinitely many
    // of u + \lambda k being wellformed.
    bool,obj:=second_dim1_wellformed(u,k); // order switched for tiny speedup
    if bool then
        return true,obj;
    end if;
    bool,obj:=first_dim1_wellformed(u,k);
    if bool then
        return true,obj;
    end if;
    // The only general thing that can cut P down to a finite size is the
    // quasismoothness of the final coordinate point (which has not been
    // included in the main monrange-loop since it is often big). Note: We
    // impose a necessary bound, not a sufficient one, for this quasismoothness,
    // so solutions gained here may still fail later.
    bool,obj:=first_qsmooth(u,k);
    if bool then
        return true,obj;
    end if;
    // The 1-d cone u + n * k resisted first analysis, so check qsmooth more.
    bool,obj:=second_qsmooth(u,k);
    if bool then
        return true,obj;
    end if;
    // Also check well-formedness of the whole 1-d family
    bool,poss:=most_not_wf(u,k);
    if bool then    // Only elts of seq poss are viable as well-formed solutions
        return true,poss;
    end if;
    // Check the 1-stratum of the 1-d family
    bool,poss:=check_1_dim_proportional_strata(u,k);
    if bool then    // Only elts of seq poss are viable as qsm solutions
        return true,poss;
    end if;
    // We've done all we can, so make the ray and return.
    return false,Polytope([u]) + Cone(k);
end function;

//////////////////////////////////////////////////

// The aim is to translate u into the strictly positive orthant using integer
// multiples of k in all possible ways. Return values are BOOL,OBJ which is
// either:
//  BOOL = true and OBJ is a finite sequence of all solutions (possibly empty);
//  BOOL = false and OBJ is a tuple <u,k> where u is now the extreme integral
//  point of an infinite half-line.
function shift_point(u,k)
    neg_indices:=[Integers() | i : i in [1..#u] | u[i] le 0];
    if #neg_indices eq 0 then
        sgn:=0;
        for i in [1..#k] do
            sgn:=Sign(k[i]);
            if sgn ne 0 then break; end if;
        end for;
        if sgn eq 0 then return true,[u]; end if;
        sgn *:= -1;
    else
        sgn:=Sign(k[neg_indices[1]]);
        if sgn eq 0 then
            return true,[PowerSequence(Integers())|];
        end if;
        for j in [2..#neg_indices] do
            if Sign(k[neg_indices[j]]) ne sgn then
                return true,[PowerSequence(Integers())|];
            end if;
        end for;
    end if;
    // Set k to be pointing in the right direction for shifting u.
    if sgn eq -1 then
        k:=[Integers() | -a : a in k];
    end if;
    // How far can we translate? We seek u + m*k with m >= 0.
    upper:=-1;
    lower:=-1;
    for i in [1..#k] do
        if u[i] gt 0 then
            if k[i] lt 0 then
                uu:=Floor((u[i] - 1) / -k[i]);
                if upper eq -1 then
                    upper:=uu;
                else
                    upper:=Minimum(upper,uu);
                end if;
            end if;
        else
            if k[i] gt 0 then
                lower:=Maximum(lower,Ceiling((-u[i] + 1) / k[i]));
            else
                assert false;	// This was handled earlier.
            end if;
        end if;
        if upper ne -1 and lower gt upper then
            return true,[PowerSequence(Integers())|];
        end if;
    end for;
    if lower ne -1 and upper ne -1 then // We know the region
        res:=[PowerSequence(Integers())|];
        for j in [lower..upper] do
            v:=[Integers() | u[i] + j * k[i] : i in [1..#u]];
            if check_1_strata(v) then
                Append(~res,v);
            end if;
        end for;
        return true,res;
    elif lower ne -1 then	// Extreme point is u + lower * k
        u:=[Integers() | u[i] + lower * k[i] : i in [1..#u]];
    elif upper ne -1 then	// Extreme point is u + upper * k
        u:=[Integers() | u[i] + upper * k[i] : i in [1..#u]];
        k:=[Integers() | -a : a in k];
    else
        assert false;
    end if;
    assert &and [a gt 0 : a in u];
    // Now repeat to see whether the region is finite or not.
    upper:=-1;
    for i in [1..#k] do
        if k[i] lt 0 then
            uu:=Floor((u[i] - 1) / -k[i]);
            if upper eq -1 then
                upper:=uu;
            else
                upper:=Minimum(upper,uu);
            end if;
        end if;
    end for;
    if upper ne -1 then	// We have a finite region
        res:=[PowerSequence(Integers())|];
        for j in [0..upper] do
            v:=[Integers() | u[i] + j * k[i] : i in [1..#u]];
            if check_1_strata(v) then
                Append(~res,v);
            end if;
        end for;
        return true,res;
    else
        return false,<u,k>;
    end if;
end function;

// Do our ad hoc mod 2 test - it eliminates all cases or none.
function first_dim1_wellformed(v,k)
    n:=#v - 1;
    vmod2:=[v[i] mod 2 : i in [1..n]];
    vdeg2:=v[#v] mod 2;
    vkmod2:=[(v[i] + k[i]) mod 2 : i in [1..n]];
    vkdeg2:=(v[#v] + k[#v]) mod 2;
    if (&+vmod2 le 1 and &+vkmod2 le 1)  // so ambient not wf
        or ((&+vmod2 eq 2 and vdeg2 eq 1) and
            (&+vkmod2 eq 2 and vkdeg2 eq 1)) then	// so hsf not wf
        // no multiple v+lambda*k is wellformed (stab Z/2)
        return true,[PowerSequence(Integers())|];
    end if;
    return false,_;
end function;

// Needs more comment - but see paper for some
function second_dim1_wellformed(v,k)
    n:=#v - 1;
    // We simply check for wellformedness.
    num_zeroes:=#[1 : i in [1..n] | v[i] eq k[i]];
    if num_zeroes ge n - 1 then
	// the ambient space for v + lambda*k not wellformed for any lambda >= 1
	    return true,[PowerSequence(Integers()) | v];
    end if;
    if num_zeroes eq n - 2 and v[#v] - k[#v] ne 0 then
        // Test for codim 1 (IN X) stabilisers - so X not wellformed
        // That happens for (v-k)+ik with i>1 not dividing (v-k)[#v]
        res:=[PowerSequence(Integers())|];
        for i in Append(Exclude(Divisors(Abs(v[#v] - k[#v])),1),0) do
            vv:=[Integers() | v[j] + i * k[j] : j in [1..#v]];
            if check_1_strata(vv) then
                Append(~res,vv);
            end if;
        end for;
        return true,res;
    end if;
    return false,_;
end function;

// Returns the sequence u of all points u = v + lambda * k, lambda in NN_0, such
// that u = (u_1,...,u_n,u_{n+1}) can be interpreted as X_{u_{n+1}} in
// P(u_1,...,u_n).
// Return EITHER true, finite seq which includes all solutions OR false,_.
// If false,_ then the 1-d ray v + n*k needs more analysis.
function first_qsmooth(v,k)
    n:=#v - 1;
    if k[n] eq 0 then
        return false,_;
    end if;
    // Dependence between lambda and N in x_n^N*x_i is a hyperbola.
    L:=0;
    for i in [1..n] do
        // Compute an upper bound on lambda for x_i to be the tangent mono
        // (the same calculation works for i = n).
        det:=v[n] * (k[n + 1] - k[i]) - k[n] * (v[n + 1] - v[i]);
        // Next few lines: throwing away excl cases which have infinitely
        // many qsmooth solutions so need sing analysis to reduce to finite.
        if det eq 0 then
            // The hyperbola is degenerate - a vertical and horizontal line.
            // N = (k_n+1 - k_i)/k_n works for all lambda - but is only
            // a solution if it is an integer.
            if IsDivisibleBy(k[n + 1] - k[i],k[n]) then
                return false,_;
            end if;
            // Otherwise no lambda works, so x_i not tangent, so continue.
            // (The vertical line of the hyperbola has coord -u_s/k_s < 0.)
        else
            if det gt 0 then
                // lambda approaches (w_n + 1 - w_i) / v_n from below, so...
                last_int:=Ceiling( (k[n + 1] - k[i]) / k[n] ) - 1;
            else
                last_int:=Floor( (k[n + 1] - k[i]) / k[n] ) + 1;
            end if;
            biggest_lambda:=Floor( (v[n + 1] - v[i] - v[n] * last_int) /
                                        (k[n] * last_int - k[n + 1] + k[i]) );
            L:=Maximum(L,biggest_lambda);
        end if;
    end for;
    res:=[PowerSequence(Integers())|];
    for lambda in [0..L] do
        vv:=[Integers() | v[j] + lambda * k[j] : j in [1..n+1]];
        if check_1_strata(vv) then
            Append(~res,vv);
        end if;
    end for;
    return true,res;
end function;

// Input: u + n*k is a 1-d family --- check qsmooth on 1-d strata
// Return EITHER true, finite seq which includes all solutions OR false,_.
// If false,_ then the 1-d ray u + N*k needs more analysis.
//
// e.g. u = [3,3,2,2,2,11], k = [2,2,1,1,0,6], then the 1-stratum L_12
// is usually not qsmooth bcs x5 is the only tangent monomial there.
// In this case we return true, [ [u] ]. (u survives by the fluke 2,2,2.)
function second_qsmooth(u,k)
    n:=#k;
    zeros_in_k:={Integers() | i : i in [1..n - 1] | k[i] eq 0};
    if #zeros_in_k ne 1 then
        return false,_;
    end if;
    c:=u[Representative(zeros_in_k)];
    // find pairs of entries of k (and in parallel u) that are equal.
    for I in Subsets({1..n - 1} diff zeros_in_k,2) do
        i1,i2:=Explode(Setseq(I));
        if k[i1] eq k[i2] and u[i1] eq u[i2] and IsDivisibleBy(k[n],k[i1])
                                          and not IsDivisibleBy(u[n],u[i1]) then
            // ... so the 1-stratum P_I = P[i1-i2] is contained in the hsf.
            // Run through the other variables xj, collecting multiples la of k
            // for which xj is a second tangent var to P_I in u+la*k for
            // infinitely many la.
            for j in {1..n - 1} diff ({i1,i2} join zeros_in_k) do
                if k[i1]*u[n] - k[i1]*u[j] - u[i1]*k[n] + u[i1]*k[j] eq 0 then
                    continue I; // not using this 1-stratum to impose conditions
                end if;
            end for;
            // If we get here, then there are only finitely many la that
            // can give a second tangent mono along P_I (other than x_zeros...).
            // Find the possibilities for each j and return them all -- after
            // all, these are the only ways THIS stratum can be qsmooth.
            poss_N:={Integers()|};
            for j in {1..n - 1} diff ({i1,i2} join zeros_in_k) do
                disc:=-k[i1] * u[n] + k[i1] * u[j] + u[i1] * k[n] - u[i1] *k[j];
                if disc gt 0 then
                    maxM:=Ceiling((k[n] - k[j]) / k[i1]) - 1;
                    for M in [1..maxM] do
                        N:=(u[n] - u[j] - M * u[i1]) / (k[j] + M *k[i1] - k[n]);
                        if IsIntegral(N) and N ge 0 then
                            Include(~poss_N,N);
                        end if;
                    end for;
                else // disc lt 0
		    // hyperbola is decreasing, and < 0 near infinity
		    // because (by this stage) both u[i1] and k[i1] > 0
		    // so horizontal asympt = - u[i1]/k[i1] < 0.
                    M:=Floor((k[n] - k[j]) / k[i1]) + 1;
                    N:=(u[n] - u[j] - M * u[i1]) / (k[j] + M * k[i1] - k[n]);
                    while N ge 0 do
                        if IsIntegral(N) then
                            Include(~poss_N,N);
                        end if;
                        M +:= 1;
                        N:=(u[n] - u[j] - M * u[i1]) / (k[j] + M *k[i1] - k[n]);
                    end while;
                end if;
            end for;
            res:=[PowerSequence(Integers())|];
            for N in poss_N do
                v:=[Integers() | u[i] + N * k[i] : i in [1..n]];
                if check_1_strata(v) then
                    Append(~res,v);
                end if;
            end for;
            return true,res;
        end if;
    end for;
    return false,_;
end function;

// If a codim 2 stratum of u[1..n-1] and k[1..n-1] are proportional, but that
// does not extend to the n-th coordinate, then that stratum is contained
// in the hypersurface with its non-trivial stabiliser.
// (See paper for bound 0 < |eps| <= alpha.)
// In this case return true.
// (So if false,_ is returned then the whole solution space is still viable,
// whereas if true,seq is returned then only elts of seq are still viable.)
function most_not_wf(u,k)
    n:=#u;
    nonzero_ki:=[Integers() | i : i in [1..n - 1] | k[i] ne 0];
    if #nonzero_ki lt n - 3 then
        return false,_;
    end if;
    for i in [1..Minimum(3,#nonzero_ki)] do
	// if you haven't found n-3 var wts in ratio by then, then you won't.
        j:=nonzero_ki[i];
        alpha:=u[j] / k[j];	// nec'y > 0 bcs k[j]<0 filtered out earlier
        eps:=u[n] - alpha * k[n];
        if eps ne 0 and #[i : i in [1..n-1] | u[i] - alpha*k[i] ne 0] le 2 then
            N:=Floor(Abs(eps) - alpha);
            res:=[PowerSequence(Integers())|];
            for la in [0..N] do
                v:=[Integers() | u[i] + la * k[i] : i in [1..#u]];
                if check_1_strata(v) then
                    Append(~res,v);
                end if;
            end for;
            return true,res;
        end if;
    end for;
    return false,_;
end function;

//////////////////////////////////////////////////////////////////////

/*
Question: when X = v + n*k  (v = (v1...;d), k = (k1...; e))
1. Is L_ij contained inside X?
2. Does X have two distinct tangent monomials along L_ij?

We impose hypotheses:
    v_i k_j - v_j k_i = 0
    ki, kj > 0
Define a positive rational number N by v_i = N * k_i.

For 2: run through each var index l (not= i,j) and
    either we give an upper bound n <= B_l
    or there is no restriction on n - treat this as n <= B_l = infinity.
Then for any n > second largest of the B_l, this stratum cannot be
quasismooth. (So if second largest B_l is infinity, the stratum is good.)

Be careful about zero denominators in the calculation - these are the only way
B_l = infinity can occur.

The sum is: xi^a xj^b xl in f (for X = v + n*k) imples
    a(vi + n*ki) + b(vj + n*kj) + (vl + n*kl) = d + n*e.

if a*ki + b*kj + kl - e = 0 and d - a*vi - b*vj - vl = 0 then
    B_l = infinity
else
    n = ( -N * (a*ki + b*kj) - (vl - d) ) / ( (a*ki + b*kj) + (kl - e) )
    so ...
end if;

so ... bounding that expression gives B_l.
That's easy:
set
    C = N (kl - e) - (vl - d)
so
    n = -N + C / ( (a*ki + b*kj) + (kl - e) ).
Then B_l = floor( max( Abs(C) - N, 0) ).

In short:
    B_l = max( Abs( N (kl - e) - (vl - d) ) - N, 0).

For 1: the same but a one-off without vl, kl in the formula.

if a*ki + b*kj - e = 0 and d - a*vi - b*vj = 0 then
    L_ij is not in X
else
    n = ( -N * (a*ki + b*kj) + d ) / ( (a*ki + b*kj) - e )
    hence
    B = floor( max( Abs( d - N*e ) - N, 0) )
    is an upper bound for those n for which L_ij is not in X.
end if;

If the relevant upper bounds are finite, then L_ij excludes all such
X when n > bound, and for n in [0..bound] return finite sequence.

*/

// return true,_ if v+n*k is quasismooth along L_I for infinitely many n
// or the input does not satisfy the test hypotheses so cannot be limited here.
// return false,[poss solutions] if only finitely many are qsmooth along L_ij.
function is_quasismooth1(I,v,k)
    v := Eltseq(v);
    k := Eltseq(k);
    n := #v;
    assert #I ge 1 and &and[ i in [1..n-1] : i in I ];
    assert #k eq n;

    // Step 1: upper bound on n for which L_ij not in X.
    if &or [ k[i] le 0 : i in I ] then
        return true,_;	// we simply don't test under these circumstances
    end if;
    // We know (from usage) that v is proportional to k for indices in I.
    N := v[I[1]]/k[I[1]];
    d := v[n];
    e := k[n];
    vI := [v[i]:i in I];
    kI := [k[i]:i in I];
    bool,sol,ker := IsConsistent(Matrix(#I,1,kI),Matrix(1,1,[e]));
    if bool and d - &+[ sol[1,i]*vI[i] : i in [1..#I] ] eq 0 then
        // The monomial x^sol prevents L_I lying in X, so no qsmooth condition.
        return true,_;
    end if;
    // Otherwise L_I can fail to lie in X sporadically, but only for lambda
    // bounded above by B: if lambda > B then L_I is contained in X so
    // needs testing for qsmoothness.
    B := Floor( Maximum( [Abs( d - N*e ) - N, 0] ) );

    // Step 2: upper bounds for n for which x_j not tangent mono along L_ij.
    bounds := [Integers()| B ];
    number_of_infinite_bounds := 0; // if this reaches #I, X qsmooth along L_I
    for l in [1..n-1] do
        if l in I then continue l; end if;
        vl := v[l];
        kl := k[l];
        bool,sol,ker := IsConsistent(Matrix(#I,1,kI),Matrix(1,1,[e-kl]));
        if bool and d - vl - &+[ sol[1,i]*vI[i] : i in [1..#I] ] eq 0 then
            // This monomial 'x_l * x_^sol' gives tangent var on the 1-stratum
            number_of_infinite_bounds +:= 1;
            if number_of_infinite_bounds ge #I then
                // enough tangent vars, so stratum L_I is fine even inside X
                return true,_;
            end if;
            continue l;
        end if;
        // Otherwise, there is an upper bound for lambda for which x_l
        // provides a tangent linear form.
        Append( ~bounds, Floor( Maximum([ Abs( d - N*e - vl + N*kl ) - N, 0]) ) );
    end for;
    // We haven't found enough monomials on this stratum that work as tangent
    // monomials for infinitely many members of the series.
    // So this stratum can only be quasismooth for finitely many members.
    // Rather than being very sharp about this bound, simply take the biggest.
    return false,Maximum(bounds);
end function;

// v = [v1,...], k = [k1,..].
// Return seqs I = [i1..ir] s.t. vi = a * ki for all i in I and fixed a=a_I > 0.
function proportional_strata(v,k)
    results := [ PowerSequence(Integers()) | ];
    v := Eltseq(v);
    k := Eltseq(k);
    n := #v;
    assert #k eq n;
    for i in [1..n-2] do  // no point checking the last weight v[n-1] separately
        if i in &cat results then continue i; end if;
        if v[i]*k[i] ne 0 then
            a := v[i] / k[i];
            I := [ j : j in [i..n-1] | v[j] eq a * k[j] ];
            if #I ge 2 then
                Append(~results,I); 
            end if;
        end if;
    end for;
    return results;
end function;

// Return false,_ if there is no bound on n for which v+n*k is 1-quasismooth.
// Return true,[poss solns] if only finitely many v+n*k are 1-qsmooth.
function check_1_dim_proportional_strata(v,k)
    upper_bound := -1;
    v := Eltseq(v);
    k := Eltseq(k);
    n := #v;
    assert #k eq n;
    risky_strata := proportional_strata(v,k);
    for I in risky_strata do
        bool,bound := is_quasismooth1(I,v,k);
        if not bool then
            if upper_bound eq -1 then
                upper_bound := bound;
            else
                upper_bound := Minimum([upper_bound,bound]);
            end if;
        end if;
    end for;
    if upper_bound eq -1 then // no bound on lambda was detected
        return false,_;
	end if;
    res:=[PowerSequence(Integers())|];
    for i in [0..upper_bound] do
        u:=[Integers() | v[j] + i * k[j] : j in [1..n]];
        if check_1_strata(u) then
            Append(~res,u);
        end if;
    end for;
	return true,res;
end function;
