//////////////////////////////////////////////////////////////////////
// dim2.m
//	Analyse certain 2-dimensional cones
//////////////////////////////////////////////////////////////////////

/*
Study the polyhedron v + <k1,k2> under hypotheses:
	v  = a1*(b+c), ..., a(n-3)*(b+c), b+c, b, an ; d
	k1 = a1, .....    , a(n-3),       1,   0, 0  ; e1 = sum
	k2 = a1, .....    , a(n-3),       1,   1, 0  ; e2 = sum
and canonical class K = O(k) where
	d = (sum^{n-3} ai*(b+c)) + (b+c) + b + an + k.
If k+an < 0 then only possible qsmooth cases are
	v + N1*k1 + N2*k2 where N1+2*N2 le -(k+an)-c.
elif k+an = 0 then only possible qsmooth cases lie in the 1-d family
	v + N*k2 for N >= 0,
	and if c=0 these fail except when N = 0, so only poss case is v.
else... [but we don't expect families in this case]
only cases are in 1-d families
	v + N1*k1 + N2*k2
	where 0 <=(strict if c=0) N1 <= k+an-c and N2 >= 0.

We may assume that v has our usual decreasing order.

Return bool, [sporadic cases], [families] as follows:
(1) if input F is not a family of this type, return false,_,_. (i.e. done nowt)
(2) otherwise return true,[sporadic],[families], either of which may be
empty, and the families need no further analysis.
*/

function two_dim_solution(F)
    // First, is F of the right type for our lemma?
    CF := InfinitePart(F);
    if Dimension(CF) ne 2 then
        return false,_,_;
    end if;
    r1,r2 := Explode(Rays(CF));
    k1 := Eltseq(r1);
    k2 := Eltseq(r2);
    n := #k1 - 1;
    if k1[n] ne 0 or k1[n-1] ne 0 or k1[n-2] ne 1 or
	    k2[n] ne 0 or k2[n-1] ne 1 or k2[n-2] ne 1 then
        return false,_,_;
    end if;
    if not &and [ k1[i] eq k2[i] : i in [1..n-3] ] then
        return false,_,_;
    end if;
    VF := Vertices(CompactPart(F));
    if #VF gt 1 then
        return false,_,_;
    end if;
    vv := VF[1];
    v := Eltseq(vv);
    b := v[n-1];		// b > 0
    c := v[n-2] - v[n-1];	// c >= 0
    d := v[n+1];
    if not &and [ v[i] eq (b+c)*k1[i] : i in [1..n-3] ] then
        return false,_,_;
    end if;
    // calc canonical class
    k := d - &+v[1..n];
    if k+v[n] lt 0 then	//  only possible qsmooth cases are
			// v + N1*k1 + N2*k2 where N1+2*N2 le -(k+v[n])-c.
	possible := [ PowerSequence(Integers()) |
		X : N1 in [0..-k-v[n]-c-2*N2], N2 in [0..Floor((-k-v[n]-c)/2)]
			| check_1_strata(X) where X := Eltseq(vv+N1*r1+N2*r2) ];
        return true, possible,[];
    elif k+v[n] eq 0 then  // only possible qsmooth cases lie in the 1-d family
			// v + N*k2 for N >= 0,
	//and if c=0 these fail except when N = 0, so only poss case is v.
	if c eq 0 then
	    return true,[v],[];
	else
	    bool,sol:=one_dim_solution(k2,v);
	    if bool then
	        return true,sol,[];
	    else
	        return true,[PowerSequence(Integers())|],[sol];
	    end if;
	end if;
    else		// only cases are in 1-d families v + N1*k1 + N2*k2
			// where 0 <=(strict if c=0) N1 <= k+v[n]-c and N2 >= 0.
        sporadic := [PowerSequence(Integers())|];
	families1d := [];
	min := c eq 0 select 1 else 0;
	for N1 in [min..k+v[n]-c] do
	    bool,sol:=one_dim_solution(k2,Eltseq(vv+N1*r1));
	    if bool then
	        sporadic cat:= sol;
	    else
		Append(~families1d,sol);
	    end if;
	end for;
	return true,sporadic,families1d;
    end if;
end function;

