//////////////////////////////////////////////////////////////////////
// dimn.m
//////////////////////////////////////////////////////////////////////

/*
Input: F a polyhedron
Each Y = [a1..an,d] in F represents a hypersurface as above.
Return:
	false,Q		if only finitely-many Y in F are quasismooth;
			Q is a sequence of these Y.
	true,_		if infinitely many Y in F are quasismooth, or
			if we can't decide.

Two examples to think about:

2-dimensional polyhedron soln with 0-dimensional finite part with one vertex:
    (14, 14,  7,  6,  1,  1, 43)
and 2-dimensional infinite part given by a cone with 2 minimal generators:
    (14, 14,  7,  6,  1,  0, 42),
    (14, 14,  7,  6,  1,  1, 43)

> F := Polytope([[14, 14,  7,  6,  1,  1, 43]]) + Cone([
    [14, 14,  7,  6,  1,  0, 42], [14, 14,  7,  6,  1,  1, 43]]);
> quasismooth_family_test(F);
false [ ]
	-- i.e. there are no quasismooth hypersurfaces in this family.

2-dimensional polyhedron soln with 0-dimensional finite part with one vertex:
    (20/3, 20/3,   4, 5/3,   1,   1,  21)
and 2-dimensional infinite part given by a cone with 2 minimal generators:
    (  20,   20,  12,   5,   3,   0,  60),
    (  20,   20,  12,   5,   3,   3,  63)

> F := Polytope([ [20/3, 20/3,   4, 5/3,   1,   1,  21] ]) + Cone([
[  20,   20,  12,   5,   3,   0,  60], [  20,   20,  12,   5,   3,   3,  63] ]);
> quasismooth_family_test(F);
false []
	-- i.e. there are no quasismooth hypersurfaces in this family.

*/

function wellformed_family_test(F)
    cpt := CompactPart(IntegralPart(F));
    points := Points(cpt);
    v := Representative(points);
    n := #Eltseq(v) - 1;
    // only work on this if all p in points have the same first n-1 coords.
    if not &and [ Eltseq(p)[1..n-1] eq Eltseq(v)[1..n-1] : p in points ] then
        return true,_;
    end if;
    zero := Zero(Parent(v));
    rays := ZGenerators(InfinitePart(F));
    if v.1 eq 0 then    // not likely, so ignore for now
        return true,_;
    end if;
    ratios := [r.1/v.1 : r in rays];
    comps := [ rays[i] - ratios[i]*p : i in [1..#rays],p in points ];
    // Testing n-1 coords checks the ambient wps is well formed.
    // Aside: we could test later for any subset of n-2 of these for
    // well-formedness of the hypersurface - but we would need to check
    // that the equation does not cut out something in codim 2 there.
    if &and [ Eltseq(s)[1..n-1] eq Eltseq(zero)[1..n-1] : s in comps ] then
        // Only the lattice points of the leading integral face have a
        // chance of being quasismooth.
        maybe := {PowerSequence(Integers())| Eltseq(p) :
                p in Points(CompactPart(IntegralPart(F)))};
        return false, [PowerSequence(Integers()) | X : X in maybe |
                            is_wellformed(X) and is_quasismooth(X) ];
    end if;
    return true,_;
end function;

function quasismooth_family_test(F)
    bool,sol := wellformed_family_test(F);
    if not bool then
        return false,sol;
    end if;
    return true,_;
end function;

