//////////////////////////////////////////////////////////////////////
// hypersurfaces.m
//
// Code for classifying well-formed, quasismooth n-fold hypersurfaces.
// X_d in P(a_1,...,a_{n+2}) s.t. sum(a_i) - d = fanoindex.
// The output is printed to the file given by filename - default="nofile".
//
// Gavin Brown, Alexander Kasprzyk 2009-15.
// Developed from an implementation by Olof Sisask in 2005 of an
// algorithm of Miles Reid first coded in 1979. Also closely related to
// an algorithm of Jennifer Johnson and Janos Kollar coded in 2000-1.

//////////////////////////////////////////////////////////////////////
// Reid's algorithm: main finiteness lemma and starting point.

// Build a matrix that records powers of variables occurring in
// those monomials that guarantee quasismoothness at the 0-strata.
// The first row is there to fix the fano index:
// M = ( 1 ... 1, -1, fanoindex) meaning that sum of wts - deg f = fanoindex.
/*
    M:=Matrix(1,n+2,[Integers()|1 : i in [1..n]] cat [-1,fanoindex]);

// We assume a_1 >= ... >= a_{n+2}.
// The mon range records possible monomials that give quasismoothness
// for the current variable under consideration (namely x1 to start with).
// The format is [5,2] indicates x1^(5-1)*x2; i.e. 5 is degree, 2 is x2=tgt var.
// Lemma: x1 cannot appear to degree > 5.
    monrange:=[
      [5,1],[5,2],[5,3],[5,4],[5,5],[5,6],
      [4,1],[4,2],[4,3],[4,4],[4,5],[4,6],
      [3,1],[3,2],[3,3],[3,4],[3,5],[3,6],
      [2,1]];
    monrange:=Reversion(monrange);
*/
//////////////////////////////////////////////////////////////////////

// Load necessary supporting files:
load "qsmooth.m";           // returns true iff general X_d is quasismooth.
load "convexgeom.m";        // finds positive solutions to inhgs linear alg.

load "../Sing/getsings.m";
load "../Sing/singtype.m";

//////////////////////////////////////////////////////////////////////
//			Main

// Input:
//     n = dimension of hypersurface X in wP^(n+1)
//     k = canonical degree, i.e. K_X = O_X(k).

// Output: set of sequences [a_1,...,a_n+1,d] rep'g X_d in P(a_1,...,a_n+1)
// with K_X=O_X(k).

forward hsfs, remove_subset_families;
function QuasismoothHypersurfaces(n, k : filename:="nofile", monrange:=[],
                                        tidyresults:=true)
    // The argument 'n' is the dimension of the resulting hypersurfaces
    // while the variable 'n' below is that + 2. So:
    n +:= 2;	// now this is the number of variables in P^(dim(X)+1)
    M:=Matrix(1,n+2,[Integers()|1 : i in [1..n]] cat [-1,-k]);
    if #monrange eq 0 then
        if k lt 0 then
            if k gt 2-n then
                monrange:=&cat[[[n-i,j] : j in [1..n]] : i in [1..n-3]] cat [[2,1]];
            else
                monrange:=&cat[[[n-i,j] : j in [1..n]] : i in [1..n-2]];
            end if;
        else
            monrange:=&cat[ [[n+k-i,j] : j in [1..n]] : i in [0..n+k-2]];
        end if;
    end if;
    Reversion(~monrange);
    range_cache:=AssociativeArray( Integers() );
    results:={PowerSequence(Integers()) | };
    families:={};	// this is for polyhedron solutions
    hsfs(M, monrange, 1, n, ~range_cache, ~results, ~families : filename:=filename);
    new_fams := [];
    for F in families do
        bool,extra := quasismooth_family_test(F);
        if not bool then
            results join:= SequenceToSet(extra);
        else
            Append(~new_fams,F);
        end if;
    end for;
    families := remove_subset_families(new_fams);
    sporadic := SetToSequence(results);
    return sporadic,families;
end function;

function Qpos(S)
    t:=0;
    for c in Reverse(S) do
        t:=Max(t + c,0);
    end for;
    return t;
end function;

function Qneg(S)
    t:=0;
    for c in Reverse(S) do
        t:=Min(t + c,0);
    end for;
    return t;
end function;

//////////////////////////////////////////////////////////////////////
//		Local functions

// n is the number of variables (so dim(X) = n - 2).

forward record_results;
procedure hsfs(matrix, monrange, var_i, n, ~range_cache, ~results, ~families :
    filename:="nofile")

    // var_i represents which variable we are dealing with, out of n
    adjoin:=ZeroMatrix(Integers(),1,n+2);
    adjoin[1,n + 1]:=-1;
    for mon in monrange do
        mi:=mon[1];
        ji:=mon[2];
        adjoin[1,var_i]:=mi - 1;
        adjoin[1,ji]:=adjoin[1,ji] + 1;
        M:=VerticalJoin(matrix, adjoin);
        adjoin[1,var_i]:=0;	// Reset adjoin ready for next loop
        adjoin[1,ji]:=0;
        N:=EchelonForm(M);

        if var_i ne n-1 then
            // The iterative loop
            bb:=N[var_i + 1,n + 2];
            cc:=-N[var_i + 1,n + 1];
		    Pjs := [ N[var_i+1,i] : i in [var_i+1..n] ];

	    // ----------------------------------------------------
            if cc lt 0 then
                // if all entries are >= 0, then we have (0...0,pi...ps,-c,b)
                // so ai*pi+..._as*ps + (-c)*deg(f) = b so
                // deg(f) <= (b-sum pj)/(-c), so no monomial can appear in it
                // with (usual) degree bigger than that: m <= (b-sum(pj))/(-c).
                // In fact, the same is true as long as the partial sums
                // pi, pi+p(i+1), pi+...+p(i+2),..., pi+...+pn >= 0.
                if &and[ &+Pjs[1..i] ge 0 : i in [1..#Pjs] ] then
                    max := Floor( (bb - &+Pjs) / (-cc) );
                elif bb ge 0 then
                    max:=Ceiling( (bb-Qneg(Pjs))/(-cc) );
				else
                    max:=Ceiling( (-Qneg(Pjs))/(-cc) );
                end if;

	    // ----------------------------------------------------
            elif cc eq 0 then
                max:=0;
                // There are certainly no results if the last line has a
                // monomial of (naive) degree greater than its required degree.
                // Otherwise, it's worth searching now using the usual solver.
                if not(&and[ &+Pjs[1..i] ge 0 : i in [1..#Pjs] ] and
                        &+Pjs gt bb) then
                    record_results(~results,~families,N,filename,n);
                end if;

	    // ----------------------------------------------------
            else
                if bb ge 0 then
                    max:=Ceiling(Qpos(Pjs) / cc);
                else
                    max:=Ceiling((Qpos(Pjs)-bb)/cc);
                end if;
            end if;
            if IsDefined(range_cache,max) then
                range:=range_cache[max];
            else
                range:=Reversion([[m,j] : j in [1..n],m in [2 .. max]]);
                range_cache[max]:=range;
            end if;
            // if last mono is mon = [m,j], can cut off range at [m-1,var_i-1].
            range := [ I : I in range | I[1] ge mon[1] or (I[1] lt mon[1] and I[2] lt small) ] where small := Minimum(var_i,mon[2]);
            if #range gt 0 then
                $$(N,range,var_i + 1,n,~range_cache,~results, ~families : filename:=filename);
            end if;
        else
            record_results(~results,~families,N,filename,n);
        end if;
    end for;
end procedure;

procedure record_results(~results,~families,N,filename,n)
    // n := Nrows(N);
    matr:=Submatrix(N,[1..Nrows(N)],[1..n + 1]);
    targ:=Submatrix(N,[1..Nrows(N)],[n + 2]);
    soln, polys := solve_system(matr,targ);
    for Y in soln do
        if &and[Y[i] gt 0 : i in [1..#Y]] and is_wellformed(Y) and
            is_quasismooth(Y) then
            if not Y in results then
                Include(~results,Y);
                if filename ne "nofile" then
                    fprintf filename,"%o\n",Y;
                end if;
                printf "Number %o: %o\n",#results,Y;
            end if;
        end if;
    end for;
    for P in polys do
        if P notin families then
            Include(~families,P);
            printf "Family %o: %o\n",#families,P;
            if filename ne "nofile" then
                fprintf filename,"%o\n",P;
            end if;
        end if;
    end for;
end procedure;

//////////////////////////////////////////////////////////////////////
// Remove families that are contained in others.

function remove_subset_families(families)
    idxs_to_remove := {};
    for i in [1..#families] do
        for j in [1..#families] do
             if j eq i then continue; end if;
             if families[i] subset families[j] then
                  Include(~idxs_to_remove,i);
             elif families[j] subset families[i] then
                  Include(~idxs_to_remove,j);
             end if;
        end for;
    end for;
    return [ families[i] : i in [1..#families] | i notin idxs_to_remove ];
end function;
