//////////////////////////////////////////////////////////////////////
// qsmooth.m
//		Check well-formedness and quasismoothness
//////////////////////////////////////////////////////////////////////
//
// We are given X = [a1,...,an,d] representing generic
// hypersurface X_d in P(a1..an).
//
// Check that X is well formed.
// Check that X is quasismooth.
//
//////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////
// Check that X is well formed
//////////////////////////////////////////////////////////////////////

function is_wellformed(X);
    n := #X-1;
    d := X[n+1];
    W := X[1..n];
    if &or [ GCD(Remove(W,i)) gt 1 : i in [1..n] ] then
        return false;
    end if;
    for j in [1..n] do
        for i in [j+1..n] do
            if d mod GCD(Remove(Remove(W,i),j)) ne 0 then
                return false;
            end if;
        end for;
    end for;
    return true;
end function;


//////////////////////////////////////////////////////////////////////
// Check that X is quasismooth
//////////////////////////////////////////////////////////////////////
/*
Implement Thm 8.1. of Fletcher.

8.1 Theorem
The general hypersurface Xd in P = P(a0,...,an) of degree d,
where n ≥ 1 is quasismooth if and only if
  either
    there exists a variable xi for some i of weight d
  or
    for every nonempty subset I = {i0,...,ik−1} of {0,...,n},
    either
      (a) there exists a monomial x_I^M = x_i0^m0 · · · x_ik-1^mk−1 of degree d,
    or
      (b) for μ = 1,...,k, there exist monomials
	x_I^Mμ x_eμ = x_i0^m0,μ...x_ik-1^mk−1,μ x_eμ
      of degree d, where {eμ} are k distinct elements.

Here (b) means there are k distinct tangent variables.

Compare with [JK] (2-3):
for every I subset {0..n} there is injection e:I -> {0..n}
and monos x_e(i) * &* x_j^m_{ij} of deg d for every i in I.
*/

/*
Input:
A = weights of wps
d = degree of hypersurface
I subset {1..#A} = indices of a stratum
S = weights of that stratum
Return: true iff there are #I distinct tangent monomials as in Thm 8.1(2b).
*/
forward knapsack;
function enough_monos(A,d,I,S)
     count := 0;
     for b in [d-A[i] : i in {1..#A} diff I] do
        count +:= knapsack(S,b) select 1 else 0;
        if count eq #I then
            return true;
        end if;
     end for;
     return false;
end function;

/*
Input:	Y = [a0,...,an,d]
Output:	true iff general hsf Y_d in P(a0..an) is quasismooth.
*/
function is_quasismooth(Y)
    n := #Y - 2;
    d := Y[n+2];
    A := Y[1..n+1];
    Sort(~A);

    // Either Thm 8.1(1) holds:
    if d in A then
        return true;
    end if;

    // or Thm 8.1(2) holds:
    knapsack_success:={PowerSet(Integers())|};
    for length in [1..#A] do
        checked_strata := { };
        for I in Subsets({1..n + 1},length) do
            S := [A[i] : i in Sort(SetToSequence(I))];
	        if not S in checked_strata then
	            Include(~checked_strata,S);
                // either 8.1(2a) or 8.1(2b)
                if not &or[J subset I : J in knapsack_success] then
                    bool:=knapsack(S,d);
                    if bool then
                        Include(~knapsack_success,I);
                    elif not enough_monos(A,d,I,S) then
	                    return false;	// qsmooth fails on this stratum.
	                end if;
                end if;
	        end if;
        end for;
    end for;
    return true;
end function;

/* Test on Johnson-Koll'ar, Thm 8, 22 sporadic log del Pezzo surfaces.
Ys := [
[1,1,1,1],
[1,1,1,2],
[1,1,2,3],
[1,2,3,5],
[1,3,5,7],
[1,3,5,8],
[2,3,5,9],
[3,3,5,5],
[3,5,7,11],
[3,5,7,14],
[3,5,11,18],
[5,14,17,21],
[5,19,27,31],
[5,19,27,50],
[7,11,27,37],
[7,11,27,44],
[9,15,17,20],
[9,15,23,23],
[11,29,39,49],
[11,49,69,128],
[13,23,35,57],
[13,35,81,128]];
assert &and[is_quasismooth(Append(Y,&+Y-1)) : Y in Ys];
*/

//////////////////////////////////////////////////////////////////////

// Returns true iff there is a linear sum a * q1 + b * q2 = n with a,b >= 0.
function has_nonneg_soln(q1,q2,n)
    // Begin by finding any solution u1 * q1 + u2 * q2 = n, where u1,u2 may be
    // negative.
    r,a,b:=XGCD(q1,q2);
    bool,s:=IsDivisibleBy(n,r);
    if not bool then
        return false;
    end if;
    u1:=a * s;
    u2:=b * s;
    // Is there anything further to do, or do we already have a non-negative
    // solution?
    if u1 ge 0 and u2 ge 0 then
        return true;
    end if;
    // Calculate the kernel k1 * q1 + k2 * q2 = 0
    s:=q2 div r;
    k1:=a * s;
    k2:=b * s - 1;
    // If k1 and k2 are both strictly positive (or negative) then we're done
    if (k1 gt 0 and k2 gt 0) or (k1 lt 0 and k2 lt 0) then
        return true;
    end if;
    // Ensure the kernel is a primitive element before continuing
    r:=GCD(k1,k2);
    k1 div:= r;
    k2 div:= r;
    // Now we try to move to a non-negative solution by adding on the
    // appropriate number of copies of the kernel
    if u1 lt 0 then
        // u1 < 0, so calculate the min copies of k1 needed to make it positive
        if k1 eq 0 then
            return false;
        elif k1 lt 0 then
            k1 *:= -1;
            k2 *:= -1;
        end if;
        m:=Ceiling(-u1 / k1);
        // What happened to u2?
        return u2 + m * k2 ge 0;
    else
        // u2 < 0 -- proceed like above...
        if k2 eq 0 then
            return false;
        elif k2 lt 0 then
            k1 *:= -1;
            k2 *:= -1;
        end if;
        m:=Ceiling(-u2 / k2);
        return u1 + m * k1 ge 0;
    end if;
end function;

function knapsack_recurse( Q, n )
    Q := [ c : c in Q | c le n ];
    if &or[IsDivisibleBy( n, c ) : c in Q] then
        return true;
    end if;
    if #Q eq 2 then
        return has_nonneg_soln(Q[1],Q[2],n);
    end if;
    first:=true;
    while #Q gt 0 do
        b := Q[1];
        Remove( ~Q, 1 );
        for i in [n div b..1 by -1] do
            if $$( Q, n - i * b ) then
                return true;
            end if;
        end for;
        if first then
            if $$( Q, n ) then
                return true;
            end if;
            first:=false;
        end if;
    end while;
    return false;
end function;

function knapsack( Q, n )
    if #Q eq 1 then
        return IsDivisibleBy(n,Q[1]);
    end if;
    if #Q eq 2 then
        return has_nonneg_soln(Q[1],Q[2],n);
    end if;
    if not IsDivisibleBy(n,GCD(Q)) then
        return false;
    end if;
    Q := Reverse(Sort(Setseq(Seqset(Q))));
    return knapsack_recurse(Q,n);
end function;

