// A copy of hypersurfaces.m that recurses only to a given depth and
// then presents the data required to continue the recursion to completion.

/*
Arguments to hpc_data(n,k,depth,basedir):
    n,k:        dim n hsfs X with K_X = O(k).
    depth:      How many 0-strata you consider: require depth <= n.
    basedir:    The base directory for the output files.
*/


load "../Core/qsmooth.m"; 		// returns true iff general X_d is quasismooth.
load "../Core/convexgeom.m";	// finds positive solutions to inhgs linear alg.

// The maximum number of jobs per file
FILE_SIZE:=750;
// The maximum size of monrange per job
MAX_MONRANGE:=500;

// The store used to process the output into files
file_store:=NewStore();

// Returns a string representation of the given sequence
function seq_to_str(Q)
    if #Q eq 0 then
        return "[]";
    end if;
    S := "[";
    for q in Q do
        if Type(q) eq SeqEnum then
            S cat:= $$(q) cat ",";
        else
            S cat:= Sprintf("%o,",q);
        end if;
    end for;
    return S[1..#S-1] cat "]";
end function;

// Initialises a new file store ready for outputting data to files
procedure init_files(basedir)
    StoreSet(file_store,"basedir",basedir);
    StoreSet(file_store,"id",1);
    StoreSet(file_store,"outfile",Sprintf("%o/1.txt",basedir));
    StoreSet(file_store,"num",0);
    StoreSet(file_store,"buffer",[]);
end procedure;

// Flushes the buffer.
procedure flush_file()
    buffer:=StoreGet(file_store,"buffer");
    if #buffer ne 0 then
        fprintf StoreGet(file_store,"outfile"),"%o\n",Join(buffer,"\n");
        StoreSet(file_store,"buffer",[]);
    end if;
end procedure;

// Flushes the buffer and moves on to a new file.
procedure new_file()
    flush_file();
    id:=StoreGet(file_store,"id") + 1;
    StoreSet(file_store,"id",id);
    StoreSet(file_store,"outfile",Sprintf("%o/%o.txt",
                                            StoreGet(file_store,"basedir"),id));
    StoreSet(file_store,"num",0);
end procedure;

// Adds the given string to the file.
procedure print_to_file(str)
    num:=StoreGet(file_store,"num");
    if num ge FILE_SIZE then
        new_file();
        num:=0;
    end if;
    buffer:=StoreGet(file_store,"buffer");
    Append(~buffer,str);
    StoreSet(file_store,"buffer",buffer);
    StoreSet(file_store,"num",num + 1);
    if #buffer ge 10000 then
        flush_file();
    end if;
end procedure;

// Add the given data to the file.
procedure data_to_file(N,monrange)
    N:=seq_to_str(RowSequence(N));
    while #monrange gt MAX_MONRANGE do
        print_to_file(Sprintf("N:%o\nrange:%o\n",
                                      N,seq_to_str(monrange[1..MAX_MONRANGE])));
        monrange:=monrange[MAX_MONRANGE + 1..#monrange];
    end while;
    if #monrange ne 0 then
        print_to_file(Sprintf("N:%o\nrange:%o\n",N,seq_to_str(monrange)));
    end if;
end procedure;

//////////////////////////////////////////////////////////////////////

forward hsfs;
procedure hpc_data(n,k,depth,basedir : monrange:=[])
    // The argument 'n' is the dimension of the resulting hypersurfaces
    // while the variable 'n' below is that + 2. So:
    depth +:= 1;    // gets number of monos right according to the def above.
    n +:= 2;	// now this is the number of variables in P^(dim(X)+1)
    M:=Matrix(1,n+2,[Integers()|1 : i in [1..n]] cat [-1,-k]);
    if #monrange eq 0 then
        if k lt 0 then
            if k gt 2-n then
                monrange:=&cat[[[n-i,j] : j in [1..n]] : i in [1..n-3]] cat [[2,1]];
            else
                monrange:=&cat[[[n-i,j] : j in [1..n]] : i in [1..n-2]];
            end if;
        else
            monrange:=&cat[ [[n+k-i,j] : j in [1..n]] : i in [0..n+k-2]];
        end if;
    end if;
    Reversion(~monrange);
    range_cache:=AssociativeArray( Integers() );
    init_files(basedir);
    hsfs(depth,M,monrange,1,n,~range_cache);
    flush_file();
end procedure;

function Qpos(S)
    t:=0;
    for c in Reverse(S) do
        t:=Max(t + c,0);
    end for;
    return t;
end function;

function Qneg(S)
    t:=0;
    for c in Reverse(S) do
        t:=Min(t + c,0);
    end for;
    return t;
end function;

// n is the number of variables (so dim(X) = n - 2).
procedure hsfs(depth, matrix, monrange, var_i, n, ~range_cache)
    if depth eq var_i then
        data_to_file(matrix,monrange);
        return;
    end if;
    // var_i represents which variable we are dealing with, out of n
    adjoin:=ZeroMatrix(Integers(),1,n+2);
    adjoin[1,n + 1]:=-1;
    for mon in monrange do
        mi:=mon[1];
        ji:=mon[2];
        adjoin[1,var_i]:=mi - 1;
        adjoin[1,ji]:=adjoin[1,ji] + 1;
        M:=VerticalJoin(matrix, adjoin);
        adjoin[1,var_i]:=0;	// Reset adjoin ready for next loop
        adjoin[1,ji]:=0;
        N:=EchelonForm(M);

        if var_i ne n-1 then
            // The iterative loop
            bb:=N[var_i + 1,n + 2];
            cc:=-N[var_i + 1,n + 1];
		    Pjs := [ N[var_i+1,i] : i in [var_i+1..n] ];

	    // ----------------------------------------------------
            if cc lt 0 then
                // if all entries are >= 0, then we have (0...0,pi...ps,-c,b)
                // so ai*pi+..._as*ps + (-c)*deg(f) = b so
                // deg(f) <= (b-sum pj)/(-c), so no monomial can appear in it
                // with (usual) degree bigger than that: m <= (b-sum(pj))/(-c).
                // In fact, the same is true as long as the partial sums
                // pi, pi+p(i+1), pi+...+p(i+2),..., pi+...+pn >= 0.
                if &and[ &+Pjs[1..i] ge 0 : i in [1..#Pjs] ] then
                    max := Floor( (bb - &+Pjs) / (-cc) );
                elif bb ge 0 then
                    max:=Ceiling( (bb-Qneg(Pjs))/(-cc) );
				else
                    max:=Ceiling( (-Qneg(Pjs))/(-cc) );
                end if;

	    // ----------------------------------------------------
            elif cc eq 0 then
                // We don't analyse this case further, because we're not
                // good at working out the next monrange.
                // Instead, simply give HPC the current 'matrix, [mon]' data.
                if not(&and[ &+Pjs[1..i] ge 0 : i in [1..#Pjs] ] and
                        &+Pjs gt bb) then
	                data_to_file(matrix,[mon]);
				end if;
                max := 0;

	    // ----------------------------------------------------
		    else
		        if bb ge 0 then
		            max:=Ceiling(Qpos(Pjs) / cc);
		        else
		            max:=Ceiling((Qpos(Pjs)-bb)/cc);
		        end if;
		    end if;
		    if IsDefined(range_cache,max) then
		        range:=range_cache[max];
		    else
		        range:=Reversion([[m,j] : j in [1..n],m in [2 .. max]]);
		        range_cache[max]:=range;
		    end if;
            // if last mono is mon = [m,j], can cut off range at [m-1,var_i-1].
            range := [ I : I in range | I[1] ge mon[1] or (I[1] lt mon[1] and I[2] lt small) ] where small := Minimum(var_i,mon[2]);
            if #range gt 0 then
                $$(depth,N,range,var_i + 1,n,~range_cache);
            end if;
        else
            error "The depth is too deep and the loop has finished!";
        end if;
    end for;
end procedure;

/////////////////////////////////////////////////////////////////////////

//dest_dir:="/Users/alex/Desktop/hpc_hypersf_data";
dest_dir:="/home/akasprzy/hpc_hypersf_data";
procedure prepare_hpc_data(n,k)
    hpc_data(n,k,n,dest_dir);
    bool,bxcar:=IsIntrinsic("Boxcar");
    if bool then
        _:=bxcar(Sprintf("HPC data prepared for n = %o, k = %o",n,k));
    end if;
end procedure;
