//////////////////////////////////////////////////////////////////////
//              Read and parse output files from HPC
//////////////////////////////////////////////////////////////////////
/*
    ** WARNING: this assumes all results will fit in memory together **

We point magma to a directory of files like
Fano4Index1_1234.txt
CY3_12345.txt
each of which has results of the following forms:
(1) sporadic results - each one is a line like
[ 1, 2, 3, 4, 5, 1234 ]
(2) families of results - each one is several lines.
(2a) 1-dimensional families look like:
Family 1: 1-dimensional polyhedron soln with 0-dimensional finite part with one
vertex:
    (3, 2, 2, 2, 8)
and 1-dimensional infinite part given by a cone with one minimal generator:
    (2, 1, 1, 0, 4)

(2b) 2-dimensional families look like:
2-dimensional polyhedron soln with 0-dimensional finite part with one vertex:
    (42, 28, 12,  1,  1,  1, 85)
and 2-dimensional infinite part given by a cone with 2 minimal generators:
    (42, 28, 12,  1,  1,  0, 84),
    (42, 28, 12,  1,  1,  1, 85)

-------------
Plan:
if not in_family and a line does not start '[', then expect
some vertex info by setting in_vertices := true.

if in_family and in_vertices and a line does not start '(', then expect
some cone info by setting in_cone := true.

if in_cone and a line starts '[', then construct and save the previous
polygon and don't forget to write the sporadic example on the current line.
-------------

Functions:

read_hpc_output("PATH_TO_OUTPUT_DIR")
returns two sequences
    sporadic, families
containing the two types of data as magma objects.

*/

Attach("buffered_reader.m");

// Returns a sequence listing the contents of the given directory. Note that
// this makes assumptions that "dir" is given in a sensible way.
function list_directory(dir)
    return Split(Pipe(Sprintf("ls \"%o\"",dir),""));
end function;

// Reads in and returns a 1-dimensional series given as a string in the form:
//      "[a1,...,an]+[k1,...,kn]"
// where the first point represents the vertex, and the second the ray.
function read_start_plus_kernel(L)
    // Split into vertex and ray
    idx:=Index(L,"+");
    error if idx eq 0,
        "read_start_plus_kernel: Expected a 1-dimensional series.";
    vert:=eval Trim(L[1..idx-1]);
    ray:=eval Trim(L[idx+1..#L]);
    // Sanity check
    error if #vert ne #ray,
        "read_start_plus_kernel: Mismatched dimensions.";
    // Return the polyhedron
    C:=Cone(ray);
    P:=Polytope([vert]);
    return P + C;
end function;

// Reads in and returns a polyhedron from the buffered reader F.
function read_polyhedron(F)
    // Are we really in a polyhedron?
    L:=Gets(F);
    error if Index(L,"polyhedron") eq 0,
        "read_polyhedron: Expected a polyhedron.";
    // First we read in the vertices of the compact part
    verts:=[PowerSequence(Rationals())|];
    in_verts:=true;
    while in_verts and not IsEmpty(F) do
        L:=Trim(Gets(F));
        if #L ne 0 then
            if Index(L,"(") eq 0 then
                in_verts:=false;
            else
                v:=eval "[Rationals() | " cat Trim(L,"(),") cat "]";
                Append(~verts,v);
            end if;
        end if;
    end while;
    // Sanity check
    error if in_verts, "read_polyhedron: Unexpected end of file.";
    error if #verts eq 0, "read_polyhedron: Failed to read any vertices.";
    d:=#Representative(verts);
    error if not &and[#v eq d : v in verts],
        "read_polyhedron: The vertices are of different dimensions.";
    // Now read in the infinite part
    rays:=[PowerSequence(Integers())|];
    in_rays:=true;
    while in_rays and not IsEmpty(F) do
        L:=Trim(Gets(F));
        if #L ne 0 then
            if Index(L,"(") eq 0 then
                Ungets(F,L);
                in_rays:=false;
            else
                v:=eval "[Integers() | " cat Trim(L,"(),") cat "]";
                Append(~rays,v);
            end if;
        end if;
    end while;
    // Sanity check
    error if #rays eq 0, "read_polyhedron: Failed to read any rays.";
    error if not &and[#v eq d : v in rays],
        "read_polyhedron: The rays are of different dimensions.";
    // Create and return the polyhedron
    return Polytope(verts) + Cone(rays);
end function;

// Given the path to a hypersurface output data file, parses the content and
// returns the sequence of sporadic results and the sequence of families.
function read_hpc_file(path)
    // Create the empty sequences of results
    sporadic:=[PowerSequence(Integers())|];
    families:=[];
    // Wrap the file in a buffered reader
    F:=NewBufferedReader(path);
    // Start reading the data in from the file
    while not IsEmpty(F) do
        L:=Trim(Gets(F));
        if #L ne 0 then
            if Index(L,"[") eq 0 then
                Ungets(F,L);
                Append(~families,read_polyhedron(F));
            elif Index(L,"+") ne 0 then
                Append(~families,read_start_plus_kernel(L));
            else
                S:=eval L;
                Append(~sporadic,S);
            end if;
        end if;
    end while;
    // Sanity check
    d:=-1;
    if #sporadic ne 0 then
        d:=#Representative(sporadic);
        error if not &and[#S eq d : S in sporadic],
            "read_hpc_file: The sporadic results are of different dimensions.";
    end if;
    if #families ne 0 then
        if d eq -1 then
            d:=Dimension(Ambient(Representative(families)));
        end if;
        error if not &and[Dimension(Ambient(P)) eq d : P in families],
            "read_hpc_file: The families are results of different dimensions.";
    end if;
    // Return the results
    return sporadic,families;
end function;

// Given the path to a directory containing hypersurface output data files,
// parses the contents of the files and returns the sequence of sporadic results
// and the sequence of families.
function read_hpc_dir(dir)
    // Create the sets that will contain the results
    all_sporadic:={PowerSequence(Integers())|};
    all_families:={};
    // Start importing the data from the files
    d:=-1;
    for file in list_directory(dir) do
        // Import the data
        printf "Importing %o...\n",file;
        sporadic,families:=read_hpc_file(Sprintf("%o/%o",dir,file));
        // Sanity check
        if #sporadic ne 0 then
            if d eq -1 then
                d:=#Representative(sporadic);
            end if;
            error if d ne #Representative(sporadic),
                "read_hpc_dir: The results are of different dimensions.";
        elif #families ne 0 then
            if d eq -1 then
                d:=Dimension(Ambient(Representative(families)));
            end if;
            error if d ne Dimension(Ambient(Representative(families))),
                "read_hpc_dir: The results are of different dimensions.";
        end if;
        // Merge the data
        all_sporadic join:= Seqset(sporadic);
        all_families join:= Seqset(families);
    end for;
    // Convert the sets into sequences and return
    all_sporadic:=SetToSequence(all_sporadic);
    Sort(~all_sporadic);
    return all_sporadic,SetToSequence(all_families);
end function;

// Post-parses to remove sporadic examples that lie in a family, and to remove
// families that are contained in other families.
function postparse_sporadics_and_families(sporadics,families)
    if #families eq 0 then
        return sporadics,families;
    end if;
    A:=AssociativeArray(Integers());
    for F in families do
        d:=Dimension(F);
        bool,S:=IsDefined(A,d);
        if bool then
            Append(~S,F);
        else
            S:=[F];
        end if;
        A[d]:=S;
    end for;
    families:=[];
    dims:=Sort(SetToSequence(Keys(A)));
    for d in dims do
        for F in A[d] do
            if not &or[F ne FF and F subset FF : FF in A[d]] and
               not &or[&or[F subset FF : FF in A[n]] : n in dims | n gt d] then
                Append(~families,F);
            end if;
        end for;
    end for;
    L:=Ambient(Representative(families));
    ChangeUniverse(~sporadics,L);
    sporadics:=[PowerSequence(Integers()) | Eltseq(W) : W in sporadics |
                                               not &or[W in F : F in families]];
    return sporadics,families;
end function;

// Given the path to a directory containing hypersurface output data files,
// parses the contents of the files and returns the sequence of sporadic results
// and the sequence of families. The data is post-parsed to remove sporadic
// examples that lie in a family, and to remove families that are contained
// in other families.
function read_hpc_dir_and_postparse(dir)
    sporadic,families:=read_hpc_dir(dir);
    sporadic,families:=postparse_sporadics_and_families(sporadic,families);
    return sporadic,families;
end function;

// Writes out the data