//////////////////////////////////////////////////////////////////////
//                  Identify orbifold singular loci
// June 2013
// GDB, AMK
//////////////////////////////////////////////////////////////////////
/*

Input:
    X   [a1,...,an,d]   representing X_d in P(W) where W = (a1..an)

Assumption: X_d is quasismooth. This is NOT checked here.

Output:
    description of the singular loci of X

Singularities are encoded as:
    < index r, transverse type [b1..bs], number N, dim D, location in P(W) >.
The location is described by a sequence of indices of the ambient stratum
that intersects (usually is contained in) X_d in this singular stratum.
The number N only applies to isolated singularities (in which case it is
the number of occurances of that quotient singularity, given quasismooth);
in other cases it is simply reported as N=1.
*/

//////////////////////////////////////////////////////////////////////
//                  Auxilliary functions
//////////////////////////////////////////////////////////////////////
/*
Here we calculate the transverse weights to the ambient stratum of index r
that is CONTAINED IN the QUASISMOOTH hypersurface X_d in P(W).

In other words, identify a single variable that could arise as a
tangent monomial somewhere on the stratum and return the remaining
variables mod r; the quasismooth assumption means that at any point
of the stratum there is always some tangent monomial, and even if
it is not the one chosen here, mod r the remaining weights are
the same as those returned here.

So the task is to find one of the weights a in W for which ja+r = d,
for some j, and then return the remaining weights mod r.

Example:
Consider index 3 stratum of X_28 in P^6( 1, 2, 3, 4, 5, 6, 7 ):
r = 3, d = 28, W = [ 1, 2, 3, 4, 5, 6, 7 ].
tangent_monos = [ 1, 7 ]
So along P(3,6) the equation of X_d eliminates x1 or x7 (given that X_d
is indeed quasismooth), so each point on the singular stratum is of type
    1/3(2,4,6,1 or 7) = 1/3(2,1,0,1).
The function returns [2,1,0,1].

Consider index 2 stratum of X_28 in P^6( 1, 2, 3, 4, 5, 6, 7 ): STOP - this
stratum is not contained in the general X_d.
(The 'contained in' condition is not checked - we only call this function
when the ambient stratum lies on X_d.)
*/

function transverse_weights(r,d,W)
    tangent_monos := [ i : i in [1..#W] | IsDivisibleBy(d-W[i],r) ];
    error if #tangent_monos eq 0, "Hypersurface is not quasismooth";
    return [ W[j] mod r : j in [1..#W] | j ne tangent_monos[1] ];
end function;

//////////////////////////////////////////////////////////////////////
/*
Detect whether there is a monomial of degree n in wps P(Q).

Input: Q seq of integers, n an integer.
Return: true if there is at least one combination of elts of Q (with repeats)
that sum to n.
*/
function exists_mono(Q,n)
    if #Q eq 1 then
        return IsDivisibleBy(n,Q[1]);
    else
        q := Q[1];
        Q1 := Remove(Q,1);
        for i in [0..n div q] do
            if $$(Q1,n-i*q) then
                return true;
            end if;
        end for;
        return false;
    end if;
end function;


//////////////////////////////////////////////////////////////////////
/*
Identify the orbifold strata of wps P(W).
Input: W a sequence of integers.
Return sequence of tuples
    < index, weights of stratum, idxs in W of stratum variables >.

Example:
> W := [1..7];
> find_hcfs(W);
[
    <2, [ 2, 4, 6 ], [ 2, 4, 6 ]>,
    <3, [ 3, 6 ], [ 3, 6 ]>,
    <4, [ 4 ], [ 4 ]>,
    <5, [ 5 ], [ 5 ]>,
    <6, [ 6 ], [ 6 ]>,
    <7, [ 7 ], [ 7 ]>
]
> W := [1,3,5,7,9,15];
> find_hcfs(W);
[
    <3, [ 3, 9, 15 ], [ 2, 5, 6 ]>,
    <5, [ 5, 15 ], [ 3, 6 ]>,
    <7, [ 7 ], [ 4 ]>,
    <9, [ 9 ], [ 5 ]>,
    <15, [ 15 ], [ 6 ]>
]
*/

// W a seq of ints
// find maxl subseqs of W with hcf = h
// return subseq, indices
function hcf_subseq(W,h)
    E := [Integers()|];
    I := [Integers()|];
    for i in [1..#W] do
        if IsDivisibleBy(W[i],h) then
            Append(~E,W[i]);
            Append(~I,i);
        end if;
    end for;
    if GCD(E) eq h then
        return E,I;
    else
        return [Integers()|], [Integers()|];
    end if;
end function;

// W a seq of ints
// find maxl subseqs of W with hcf > 1
// return [ <h, subseq, indices> , ... ]
function find_hcfs(W)
    poss_hcfs := &join{ SequenceToSet(Divisors(d)) : d in W };
    Exclude(~poss_hcfs,1);
    poss_hcfs := SetToSequence(poss_hcfs);
    Sort(~poss_hcfs);
    results := [];
    for h in poss_hcfs do
        E,I := hcf_subseq(W,h);
        if #E ne 0 then
            Append(~results,<h,E,I>);
        end if;
    end for;
    return results;
end function;


//////////////////////////////////////////////////////////////////////
//                  Main routine
//////////////////////////////////////////////////////////////////////
/*
Input:
    X   [a1,..,an,d]
Return:
    sequence of singular strata of X_d, encoded as above.

Example: X_28 in P(1,2,3,4,5,6,7)
The ambient space has several orbifold strata (determined by common factors
among the weights). The intersection of each one with X must be determined;
here are some samples before the complete result in code:

Stratum [2,4,6]:
    there is a monomial in these vars, so X_d cuts this stratum in codim 1
    with transverse type 1/2(0,1,3,5,7) at each point; in other words, this
    1-dim'l orbifold stratum has transverse type 1/2(1,3,5,7) = 1/2(1,1,1,1).

Stratum [7]:
    there is a monomial x7^4 in the equation so X misses this ambient stratum.

Stratum [6]:
    variable of weight 4 is a tangent monomial to this stratum, so there
    is a single point of type 1/6(1,2,3,5,7) = 1/6(1,1,2,3,5).

> singularities([1,2,3,4,5,6,7,28]);
[
    <2, [ 1, 1, 1, 1 ], 1, 1, [ 2, 4, 6 ]>,
    <3, [ 1, 1, 2, 2 ], 1, 1, [ 3, 6 ]>,
    <5, [ 1, 1, 2, 2, 4 ], 1, 0, [ 5 ]>,
    <6, [ 1, 1, 2, 3, 5 ], 1, 0, [ 6 ]>
]

Also try:
> singularities([1,2,3,4,5,6,7,27]);
[
    <2, [ 1, 1, 1 ], 1, 2, [ 2, 4, 6 ]>,
    <3, [ 1, 1, 1, 2, 2 ], 4, 0, [ 3, 6 ]>,
    <4, [ 1, 1, 2, 2, 3 ], 1, 0, [ 4 ]>,
    <5, [ 1, 1, 2, 3, 4 ], 1, 0, [ 5 ]>,
    <6, [ 1, 1, 2, 4, 5 ], 1, 0, [ 6 ]>,
    <7, [ 1, 2, 3, 4, 5 ], 1, 0, [ 7 ]>
]
*/
function singularities(X)
    n := #X - 1;
    d := X[n+1];
    W := X[1..n];
    dimX := n - 2;
    strata := find_hcfs(W);
    sings := [];
    for S in strata do
        r,P,I := Explode(S);
        // Sing is index r inside space of wts P (whose vars have idx in I).
        // We study P intersection X: either X cuts out codim 1 on P or
        // X contains P.
        // In the latter case, since X is quasismooth we know that the
        // equation will eliminate one of the weights transverse to P.
        codim := n - #P;    // codim of ambient stratum in wps.
        if exists_mono(P,d) then // X cuts out codim 1 on P
            dim := dimX - codim;
            transverse := Sort([ W[i] mod r : i in [1..#W] | i notin I ]);
            if dim eq 0 then // count the number of singular points
                // Use Fletcher 9.4 rescaled in the big stratum
                mult := Floor(d*r/(P[1]*P[2]));
            else
                mult := 1;
            end if;
            sing := < r, transverse, mult, dim, I >;
        else // P contained in X
            dim := dimX - codim + 1;
            pretrans := [ W[i] : i in [1..#W] | i notin I ];
            transverse := Sort(transverse_weights(r,d,pretrans));
            sing := < r, transverse, 1, dim, I >;
        end if;
        if sing[4] ge 0 then // negative dimension means X misses stratum P.
            Append(~sings,sing);
        end if;
    end for;
    return sings;
end function;

