//////////////////////////////////////////////////////////////////////
//	Recognise terminal and canonical quotient singularities
//////////////////////////////////////////////////////////////////////
//
// X = [a1,...,an,d] is a seq representing the family of n-2-fold hsfs
//          X_d in P(a1..an).
//
//  ***  X IS ASSUMED TO HAVE A QUASISMOOTH GENERAL ELEMENT  ***
//
// We can test for 'terminal' or 'canonical' singularities, and isolated sings.
// 
// Singularities of the general X are recorded in the format:
//	< index r, [local pol'g wts], mult (if isolated), dim, subspace of P >.
//
// The parameter 'type' is string that is a comma-separated list of
// words from: "terminal", "canonical", "isolated".
//
// Return true if the general element of this family has at worst
// singularities from the class 'type'.
// Otherwise return false.
//
//////////////////////////////////////////////////////////////////////

function check_singularities(X : type:="")
    n := #X-1;
    dimX := n-2;
    d := X[n+1];
    W := X[1..n];
    sings := singularities(X);
    known_sings := AssociativeArray(PowerSequence(Integers()));
    for T in Split(type,",") do	// do easy tests (e.g. isol) first
        if T eq "isolated" then // p[4] is the dimension of singular stratum p.
            if &or [ p[4] ne 0 : p in sings ] then
                return false;
            end if;
        elif T eq "terminal" then
            if &or [p[4] ge dimX-2:p in sings] then
                return false;   // because terminal sings are in codim 3
            end if;
            for p in sings do   // impose Tai criterion
                r := p[1];
                A := p[2];
                if &or[ &+[Integers()|
                        (i*a) mod r : a in A] le r : i in [1..r-1] ] then
                    return false;
                end if;
            end for;
        elif T eq "canonical" then  // impose Tai criterion
            for p in sings do
                r := p[1];
                A := p[2];  // sing is 1/r(A)
                if &or[ &+[Integers()|
                        (i*a) mod r : a in A] lt r : i in [1..r-1] ] then
                    return false;
                end if;
            end for;
        elif T eq "crepant" then
            for p in sings do
                r := p[1];
                A := p[2];  // sing is 1/r(A)
		if #A le 3 then
		    if (&+A mod r) ne 0 then
		        return false;
		    end if;
		else
	            // I don't know an algorithm dim > 3 so use toric geom
	            // and cache the results.
		    rA := [r] cat A;
		    if IsDefined(known_sings,rA) then
		        bool := known_sings[rA];
		    else
                        L := ToricLattice(#A);
                        _,phi := AddVectorToLattice(L![a/r:a in A]);
                        C := Cone([phi(e):e in Rays(PositiveQuadrant(L))]);
                        F:=Fan(Resolution(ToricVariety(Rationals(),Fan(C))));
                        union := Polytope(&cat[[r:r in Rays(C)]:C in Cones(F)]);
		        bool := Dimension(union) lt #A;
				// resolution is crepant iff bool
			known_sings[rA] := bool;
		    end if;
                    if not bool then
                        return false;
                    end if;
		end if;
            end for;
        elif T in {"smooth","nonsingular"} then
	    return #sings eq 0;
        end if;
    end for;
    return true;
end function;


//////////////////////////////////////////////////////////////////////
// A version of the same analysis for all members of a sequence [X1,X2,...].
// Aiming to get advantage from caching of singularity data.
//////////////////////////////////////////////////////////////////////

// Q a sequence of hypersurfaces X = [a0..an,d] as above.
// Return the subsequence that satisfy check_singularities for given type.
function seq_check_singularities(Q : type:="")
    if #Q eq 0 then
        return Q;
    end if;
    type:=Split(type,",");
    results:=[Universe(Q)|];
    are_terminal:=AssociativeArray(PowerSequence(Integers()));
    are_canonical:=AssociativeArray(PowerSequence(Integers()));
    are_crepant:=AssociativeArray(PowerSequence(Integers()));
    X:=Q[1];
    n:=#X - 1;
    dimX:=n - 2;
count := 0;
    for X in Q do
count+:=1;
print count;
        d:=X[n + 1];
        W:=X[1..n];
        sings:=singularities(X);
        for T in type do

	    //			CASE: ISOLATED SINGULARITIES
            if T eq "isolated" then
                // p[4] is the dimension of singular stratum p.
                if &or[p[4] ne 0 : p in sings] then // X is has nonisolated
						// sings - forget it
                    continue X;
                end if;

	    //			CASE: TERMINAL SINGULARITIES
            elif T eq "terminal" then  // impose Tai criterion
                if &or[p[4] ge dimX - 2 : p in sings] then
                    continue X;	// because terminal sings are in codim 3
                end if;
                for p in sings do   // impose Tai criterion
                    r:=p[1];
                    A:=p[2];  // sing is 1/r(A)
                    rA:=[r] cat A;
                    bool2,bool:=IsDefined(are_terminal,rA);
                    if not bool2 then
                        bool:=&and[&+[Integers() | (i*a) mod r : a in A] gt r :
                                                               i in [1..r - 1]];
                        are_terminal[rA]:=bool;
                    end if;
                    if not bool then // X has the wrong sings - forget it
                        continue X;
                    end if;
                end for;

	    //			CASE: CANONICAL SINGULARITIES
            elif T eq "canonical" then  // impose Tai criterion
                for p in sings do
                    r := p[1];
                    A := p[2];  // sing is 1/r(A)
                    rA:=[r] cat A;
                    bool2,bool:=IsDefined(are_canonical,rA);
                    if not bool2 then
                        bool:=&and[&+[Integers()| (i*a) mod r : a in A] ge r :
			                                      i in [1..r-1] ];
                        are_canonical[rA]:=bool;
                    end if;
                    if not bool then // X has the wrong sings - forget it
                        continue X;
                    end if;
                end for;

	    //			CASE: CREPANT CANONICAL SINGULARITIES
            elif T eq "crepant" then
		// N.B. This parameter 'crepant' is experimental. It does
		// not check that the resolution chosen by the toric code
		// really is crepant in the case that there exists a crepant 
		// resolution. Use with care.
                for p in sings do
                    r := p[1];
                    A := p[2];  // sing is 1/r(A)
                    if #A le 3 then
                        if (&+A mod r) ne 0 then
                        return false;
                        end if;
                    end if;
                    rA:=[r] cat A;
                    bool2,bool:=IsDefined(are_crepant,rA);
                    if not bool2 then
                        // Have to work out the answer: bool:=true if yes.
                        // I don't know an algorithm dim > 3 so use toric geom
                        // and cache the results.
                        L := ToricLattice(#A);
                        _,phi := AddVectorToLattice(L![a/r:a in A]);
                        C := Cone([phi(e):e in Rays(PositiveQuadrant(L))]);
                        F:=Fan(Resolution(ToricVariety(Rationals(),Fan(C))));
                        union := Polytope(&cat[[r:r in Rays(C)]:C in Cones(F)]);
                        bool := Dimension(union) lt #A;
                            // resolution is crepant iff bool
                        are_crepant[rA] := bool;
                    end if;
                    if not bool then // X has the wrong sings - forget it
                        continue X;
                    end if;
                end for;

	    //			CASE: NONSINGULAR
            elif T eq "smooth" or T eq "nonsingular" then
                if #sings ne 0 then // X is singular - forget it
		    continue X;
		end if;
            end if;
        end for;
        Append(~results,X);
    end for;
    return results;
end function;

